use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub psf_parse {

    # Purpose: processes the output file from program PSF to determine
    # the resolutions where the Fourier shell coefficient (FSC) drops
    # below two specified levels. These are used to determine the
    # resolution at this stage of reconstruction and the suggested
    # high resolution to use in the next iteration. Since the FSC is
    # calculated at discrete points, we err on the side of caution and
    # choose the highest resolution BEFORE the cutoff is reached.
    #
    # Input parameters:
    #   $fsc_hicut - upper FSC cutoff
    #   $fsc_locut - lower FSC cutoff
    #   $infile    - output file from PSF
    #
    # Return values:
    #   $res_hicut - resolution where FSC drops below $fsc_hicut
    #   $res_locut - resolution where FSC drops below $fsc_locut
    #   $reached_locut  - set to 1 if FSC drops below $fsc_locut,
    #   $reached_hicut  - set to 1 if FSC drops below $fsc_hicut,

    my ($fsc_hicut, $fsc_locut, $infile) = @_;

    my $date = localtime();

    print OUT "Parsing PSF output file\n";
    print OUT "   Date = $date\n";

    # Define flags
    my $reached_hicut = 0;  # FSC has not reached upper cutoff (0.5 default)
    my $reached_locut = 0;  # FSC has not reached lower cutoff (0.3 default)

    # Define resolutions
    my $res_hicut; # Resolution at the higher FSC cutoff
    my $res_locut; # Resolution at the lower FSC cutoff

    # Define variable holding resolution from previous record
    my $prev_res = 1000; # Initialize to ridiculously high value

    # Initialize accumulator for sum over FSC
    my $fsc_sum = 0.0;
    
    open(IN, $infile);
    while(<IN>) {
	
	# Get rid of leading and trailing whitespace
	s/^\s+//;
	s/\s+$//;
	
	# Determine whether or not we reached the resolution data
	# Lines should split on whitespace into 7 numeric fields
	my ($bin, $res, $obs, $r_sf, $fsc,
	    $r_phase, $phase_diff, $extra) = split(/\s+/, $_); 
	
	unless(defined $phase_diff && $bin =~ /^\d+$/) {
	    next; # Not in data section - not enough fields
	}
	unless($bin =~ /^\d+$/) {
	    next; # Not in data section - first field not numeric
	}
	if(defined $extra) {
	    next; # Not in data section - too many fields
	}

	# Accumulate sum over FSC
	$fsc_sum += $fsc;

	# Identify the point where CC drops below high cutoff
	unless($reached_hicut) {
	    if($fsc < $fsc_hicut) {
		$res_hicut = $prev_res;
		$reached_hicut = 1;
	    }
	}
	
	# Identify the point where CC drops below low cutoff
	unless($reached_locut) {
	    if($fsc < $fsc_locut) {
		$res_locut = $prev_res;
		$reached_locut = 1;
	    }
	}

	# Save the resolution from this record to use in next iteration
	$prev_res = $res;
    }

    # If CC never drops below $fsc_locut, set $res_locut to the
    # highest resolution evaluated in PSF. Do same in the event that
    # CC never drops below $fsc_hicut.
    unless($reached_locut) {
	print OUT "  Note: FSC never drops below $fsc_locut\n";
	print OUT "        Using highest resolution evaluated in PSF\n\n";
	$res_locut = $prev_res;
    }
    unless($reached_hicut) {
	print OUT "  Note: FSC never drops below $fsc_hicut\n";
	print OUT "        Using highest resolution evaluated in PSF\n\n";
	$res_hicut = $prev_res;
    }
    
    print OUT "   Resolution where FSC drops below $fsc_hicut: $res_hicut\n";
    print OUT "   Resolution where FSC drops below $fsc_locut: $res_locut\n";
    print OUT "   Sum over FSC: $fsc_sum\n";
    print OUT "\n";
    
    close(IN);

    return($res_hicut, $res_locut, $reached_locut, $reached_hicut);
}
1;
