      PROGRAM P3DR
C     Parallel 3D Cartesian Reconstruction program
C                                                     MAIN PROGRAM: P3DR
C **********************************************************************
C *                                                                    *
C * Main routine: P3DR in file aamain.f                                *
C * Author: Robert E. Lynch (rel@cs.purdue.edu) and                    *
C *         Hong Lin (linh@cs.purdue.edu)                              *
C *         Yongchang Ji (yji@cs.ucf.edu)                              *
C * Date:   1997 -- 2002                                               *
C * Function: Main program for cartesian reconstruction                *
C * Algorithm: P3DR uses a set of pixel frames of projections of       *
C *   an object to reconstruct the 3D density map of the object.       *
C *   Firstly, a 2D Fourier transform is carried out on each           *
C *   pixel frame; then the 2D DFT is used in estimating 3D DFT        *
C *   of the object; finally, a reverse Fourier transform is           *
C *   carried out on the 3D DFT to get the density values of the       *
C *   object.                                                          *
C *      P3DR reads in orientations and pixel frames from two          *
C *      files with extension .dat and .pif and carries out            *
C *      reconstruction based on these data.                           *
C * Modification history:                                              *
C *   - Robert E. Lynch, Version 0.0, 1997                           *
C *   - Hong Lin, Version 0.1, Apr.20, 1998                          *
C *   - Hong Lin, Version 0.6, Jul.6, 1998                           *
C *   - Hong Lin, Version 1.0, Aug.18, 1998                          *
C *   - Hong Lin, Version 2.0, Sep.29, 1999.                         *
C *   - Hong Lin, Version 2.9, June 17, 2000                           *
C *   - R.E. Lynch, Aug 2000                                           *
C *        Program name set to P3DR                                    *
C *        Moved print of time and memory-use to subroutine MEM_TIME   *
C *        from main routine --- THIS DOESN"T WORK !!!!!!              *
C *        The setting of the value of  diameter  was removed --- the  *
C *          number of angstroms per pixels is not known until the     *
C *          header of the orientation file is read.                   *
C *        Used new variable  interpolants_size  to store the size     *
C *          of the array  interpolants  which is passed to several    *
C *          routines.  Simplified the routine  intlz_interp           *
C *                                                                    *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *
C *                                                                    *
C **********************************************************************

C confine_3D -- When confine_3D is .true., the edge-length of 3D
C   DFT is set to the same as the edge-length of pixel frames. 
C 
      implicit none

      include 'mpif.h'
      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'
      include '../include/mem_time.inc'
      
      logical :: confine_3D = .false.

C Local variables, used only in the main routine
      integer first_iq, last_iq, max_iq_slab,
     +     first_z,last_z, first_hk,last_hk, max_hk_slab, num_hkl,
     +     den_min_x, den_max_x, den_min_y, den_max_y, 
     +     den_min_z, den_max_z, mem_size, 
     +     min_xyz, max_xyz, hkl_max, max_z_slab,
     +     first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab,
     +     interpolants_size
      
      integer*8 clock1, clock2, clockmax, clockrate, ticks
      integer alloc_stat, algorithm
      character(len=1) precision

C     For FFTSYNTH
      integer nrffts_1_row, nrffts_1_col, nrffts_1_omega,
     +     nrffts_2_row, nrffts_2_col, nrffts_2_omega
      
      real, allocatable ::  f_vru(:), f_vru_tmp(:), F_XK_dZ(:),
     +     rho(:,:,:), density_map(:,:,:), ctf_val(:,:), abs_ctf(:,:)
      
#ifdef SPREC
      real, allocatable :: interpolants(:)
      real, allocatable :: f_hkl_ref(:)
#else
      double precision, allocatable :: interpolants(:)
      double precision, allocatable :: f_hkl_ref(:)
#endif

      complex, allocatable :: f_hkl(:)
      
      character(len=18) :: subname = 'P3DR main routine'
      
      integer i, first_indx, total_views
      REAL MAP_MIN, MAP_MAX
      real map_min_local, map_max_local
      real    enlargefact, imgprjbig, imgprjstd

C *********************************************************************
C *                      Initialization                               *
C *********************************************************************

C                                                         INITIALIZE MPI
C                                                         **************
      call MPI_INIT ( mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_INIT')
      call MPI_COMM_RANK ( MPI_COMM_WORLD, iam, mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_RANK')
      call MPI_COMM_SIZE ( MPI_COMM_WORLD, num_nodes, mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_SIZE')
      if (num_nodes > max_numnodes) then
         write(stdoutput, 9001) num_nodes, max_numnodes
 9001    format(//' Requested number of nodes = NUM_NODES    = ',i10/
     +        ' Available number of nodes = MAX_NUMNODES = ',i10/)
         call error_stop(progname, subname,  
     +        'NUM_NODES requested exceeds MAX_NUMNODES')
      endif
      
      progname = 'P3DR'
      
C     By default, use algorithm 1 and double precision
      algorithm = 1
      precision = 'd'

#ifdef ALG3
      algorithm = 3
#endif
      
#ifdef SPREC
      precision = 's'
#endif      
      
      call system_clock(count_max=clockmax, count_rate=clockrate)
      
      if(iam == 0) then
         write(6,*) 'Version RSS 15:05'
         write(6,*) 'algorithm = ', algorithm
         write(6,*) 'precision = ', precision
         call usage(algorithm, precision)
      endif

      call system_clock(clock1)

      if( iam == 0 ) then
         print *, ' Number of Nodes =', num_nodes
C     Read User's Control-Input
C     I have change the control_input to multi files.
         call info
      endif

      if( num_nodes > 1 ) then
C     CALL BCAST_PARAMETERS
C     *******************************
C     Broadcast User's parameters and PIF File Header to other nodes.
         call bcast_parameters 
      endif

C     CALL INTLZ_PARAMS
C     *****************
      call intlz_params(hkl_max, num_hkl, min_xyz, max_xyz, 
     +     nrffta_row, nrffta_col, nrffta_omega, nfc_2d_row, 
     +     nfc_2d_col, nrffts_1_row, nrffts_1_col, 
     +     nrffts_1_omega, nrffts_2_row, nrffts_2_col, 
     +     nrffts_2_omega, den_min_x, den_max_x, den_min_y, 
     +     den_max_y, den_min_z, den_max_z,
     +     confine_3D)
      
Cji 01/06/01 added 2 print statements after call of intlz_params
      if( iam == 0 ) then
         print *, ' zero_fill after  call of INTLZ_PARAMS =',zero_fill
         print *, ' nrffta_row after call of INTLZ_PARAMS =',nrffta_row
      endif

      call intlz_arrays(hkl_max, min_xyz, max_xyz, min_z, max_z,
     +     first_hk, last_hk, max_hk_slab, first_iq, last_iq,
     +     max_iq_slab, first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab, first_z, last_z, max_z_slab,
     +     npix_row)

C ---------- End of Initializationl --- Synchronize all the nodes ------

      call mpi_barrier(mpi_comm_world, mpi_err)

      call system_clock(clock2)
      ticks = clock2-clock1
      if(ticks < 0) then
         ticks = clockmax + ticks
      endif
      t_init = dble(ticks)/dble(clockrate)
      
C **********************************************************************
C *             2D DFT and estimating 3D DFT from 2D DFT               *
C **********************************************************************
C
      call system_clock(clock1)

      if(algorithm == 1) then
         interpolants_size = 
     +        (hkl_max +1)*((2*hkl_max + 1)**2) * 3
      else if(algorithm == 3) then
         interpolants_size = 3 * ( last_iq - first_iq + 1 ) 
     +        * (( 2*hkl_max + 1 )**2)
      else
         call error_stop(progname, subname, 'Invalid algorithm choice')
      endif
      
      allocate(ctf_val(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col), 
     +     abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     interpolants(interpolants_size),
     +     stat = alloc_stat)
      
      if(alloc_stat > 0) then
         call error_stop (progname, subname,  
     +        'Memory allocation ctf_val, abs_ctf, interpolants failed')
      endif 

      if(precision == 'd') then
         mem_interpolants = dble( 8 * interpolants_size ) 
      else
         mem_interpolants = dble( 4 * interpolants_size ) 
      endif
      
      mem_size = (nfc_2d_row +1) * (nfc_2d_col * 2 +1)
      mem_ctf_val = dble(8 * mem_size)

      mem_total = mem_total + mem_ctf_val + mem_interpolants 
      mem_peak  = max(mem_peak, mem_total)

C     Zero interpolants array
      interpolants = 0.0        ! Fortran 90 array syntax

      t_analy         = 0
      t_fftanaly      = 0
      total_views     = 0
      mem_cmpt_intrps = 0
      

C     Loop over particle image files
      do i = 1, nfiles
         npix_row         =  NBOX(i)
         npix_col         =  NBOX(i)
         max_views        =  NIDS(i)
         angstrom_per_pix = PIXSIZS(i)
         
C     Root process opens image file and determines enlargefact
         if (iam == 0) then
            pix_file_name = img_fnames(i)
            ort_file_name = param_fnames(i)
            call PIF_OPEN(pix_file_name, pix_file_unit, 'r')

            call get_1st_ortid(first_indx, ort_file_name)
            if(first_indx == 0) then
               print *, 'Parameter file empty or malformed'
               print *, 'Skip to next file'
            else
               call read_1_pif(first_indx, imgprjbig) 
               if(i == 1) then
                  imgprjstd   = imgprjbig
                  enlargefact = 1. 
               else 
                  enlargefact = imgprjstd / imgprjbig
               endif
               print *, 'enlargefact=',enlargefact, imgprjbig, imgprjstd
            endif
         endif

C     Skipping to next file if parameter file empty or malformed
         call mpi_bcast(first_indx, 1, mpi_integer, 0, 
     +        mpi_comm_world, mpi_err)
         if(first_indx == 0) cycle

         call mpi_bcast(enlargefact, 1, mpi_real,
     +        0, mpi_comm_world, mpi_err)
         if(mpi_err .ne. 0) then
            call error_stop(progname,subname,
     +           'Error in calling MPI_BCAST enlargefact')
         endif
         
C     Initialize CTF arrays
         ctf_val = 0.0          ! Fortran 90 array syntax
         abs_ctf = 1.0          ! Fortran 90 array syntax
         
C     Calculate CTF corrections
         if(ctfmode > 0 .and. ctfmode <= 4)  then
            call ctf_para(ctf_val, abs_ctf, 
     $           0, nfc_2d_row, -nfc_2d_col, nfc_2d_col, .true.,
     +           CTFMODE, Filter, TEMPFAC, CTF_FF1, CTF_FF2,
     +           PIXSIZS(i), units, VOLTS(i), DELF_MAJS(i),
     +           DELF_MINS(i), ANG_MAJS(i), AMPFACS(i), CS_COEFFS(i), 
     +           enlargefact, res_max, res_min)
         endif

C     Calcualate contributions of images in file i to 3DFT
         mem_cmpt = 0
         call cmpt_intrps(hkl_max,
     +        interpolants, interpolants_size,
     +        t_analy, t_fftanaly, max_iq_slab, 
     +        mem_cmpt, total_views, ctf_val, abs_ctf, algorithm)
         
         mem_cmpt_intrps = max(mem_cmpt_intrps, mem_cmpt)
         mem_peak  = max(mem_peak, mem_total+mem_cmpt)
         
         if(iam == 0) then
            call pif_close (pix_file_unit)
            print *, 'total_view = ', total_views
         endif
      enddo

      if(total_views == 0) then
         call error_stop('P3DR', 'main program', 
     +        'All particle parameter files are empty')
      endif

      deallocate(ctf_val, abs_ctf)
      mem_total = mem_total - mem_ctf_val
      
      call mpi_barrier(mpi_comm_world, mpi_err)

      call system_clock(clock2)
      ticks = clock2-clock1
      if(ticks < 0) then
         ticks = clockmax + ticks
      endif
      t_interp = dble(ticks)/dble(clockrate) - t_fftanaly - t_analy
C
C **********************************************************************
C *        Combining estimates of 3D DFT along three directions        *
C *        arrange the data before 3D DFT                              * 
C **********************************************************************
C 
      call system_clock(clock1)

C     Merging versions 1 and 3 of code
C     Version 1 with 1 proc is same as version 3

      if(algorithm == 1 .and. num_nodes > 1) then
         mem_size    = (num_hkl**2) * (-first_iq+last_iq+1) * 3
         if(mem_size < 0) mem_size = 0
         allocate(f_hkl_ref(mem_size), stat = alloc_stat)
         if (alloc_stat > 0 ) 
     +        call error_stop(progname, subname,
     +        'Error in allocating memory for F_HKL_ref')
         
         if(precision == 'd') then
            mem_f_hkl_ref = dble(8 * mem_size)
         else 
            mem_f_hkl_ref = dble(4 * mem_size)
         endif
         
         mem_total = mem_total + mem_f_hkl_ref
         mem_peak  = max(mem_peak, mem_total)
         call exch_intp(hkl_max, interpolants, f_hkl_ref)
         deallocate (interpolants)
         mem_total = mem_total - mem_interpolants
         mem_size    = (num_hkl**2) * (-first_iq+last_iq+1)
         if(mem_size < 0) mem_size = 0
         allocate(f_hkl(mem_size), stat = alloc_stat)
         if (alloc_stat > 0 )
     +        call error_stop(progname, subname,
     +        'Error in allocating memory for F_HKL')
         mem_f_hkl = dble(8 * mem_size)
         mem_total = mem_total + mem_f_hkl
         mem_peak  = max(mem_peak, mem_total)
         call realtocomplx(hkl_max, first_iq, last_iq, 
     +        f_hkl_ref, f_hkl, algorithm)
         deallocate (f_hkl_ref)
         mem_total = mem_total - mem_f_hkl_ref
      else
         mem_size = (num_hkl**2) * (-first_iq+last_iq+1)
         if(mem_size < 0) mem_size = 0
         allocate(f_hkl(mem_size), stat = alloc_stat)
         if (alloc_stat > 0) call error_stop(progname, subname,
     +        'Error in allocating memory for F_HKL')
         mem_f_hkl = dble(8 * mem_size)
         mem_total = mem_total + mem_f_hkl
         mem_peak  = max(mem_peak, mem_total)
         call realtocomplx(hkl_max, first_iq, last_iq, 
     +        interpolants, f_hkl, algorithm)
         deallocate(interpolants)
         mem_total = mem_total - mem_interpolants
      endif

C ----- This is the end of combining 3D DFT along 3 directions -----
C ----- Before move onto the next step, synchronize all the nodes ------
C
      call mpi_barrier(mpi_comm_world, mpi_err)
      
      call system_clock(clock2)
      ticks = clock2-clock1
      if(ticks < 0) then
         ticks = clockmax + ticks
      endif
      t_solvesys= dble(ticks)/dble(clockrate)
      
C     ********************************************************************
C     *                    2D Fourier synthesis                          *
C     ********************************************************************

      call system_clock(clock1)

C     Allocate memory
      mem_size = ((-min_xyz+max_xyz+1)**2) *
     +     (-first_rr_out+last_rr_out+1)
      if(mem_size < 0) mem_size = 0
      allocate(f_vru(mem_size), stat = alloc_stat)
      if (alloc_stat > 0) call error_stop(progname, subname, 
     +     'Error in allocating memory for F_VRU')
      mem_f_vru = 4 * mem_size
      mem_total = mem_total + mem_f_vru
      mem_peak  = max(mem_peak, mem_total)
      
C     FFTSYNTH_1
C     **********
      call fftsynth_1_m_slab(hkl_max,
     +     first_iq, last_iq, first_rr_out, last_rr_out,
     +     f_hkl, nrffts_1_row, nrffts_1_col, nrffts_1_omega, 
     +     min_x, max_x, min_z, max_z, 
     +     min_xyz, max_xyz, f_vru, mem_fftsynth_1_m )
      
      deallocate(f_hkl)
      mem_peak = max(mem_peak, mem_total+mem_fftsynth_1_m)
      mem_total = mem_total - mem_f_hkl
      
C ------------ This is the end of 2D Fourier synthesis -----------------
C ------ Before move onto the next step, synchronize all the nodes -----

      call mpi_barrier(mpi_comm_world, mpi_err)

      call system_clock(clock2)
      ticks = clock2-clock1
      if(ticks < 0) then
         ticks = clockmax + ticks
      endif
      t_fftsynth_1 = dble(ticks)/dble(clockrate)

C **********************************************************************
C *           Exchanging data for 1D Fourier synthesis                 *
C **********************************************************************

      call system_clock(clock1)

C     Allocate storage for subroutine EXCHANGE_2
      mem_size = (-min_xyz+max_xyz+1) * (-min_xyz+max_xyz+1) *
     +     (-frst_lst_U(1,iam)+frst_lst_U(2,iam)+1)
      allocate(F_VRU_TMP(mem_size), stat = alloc_stat)
      if (alloc_stat > 0) call error_stop(progname,
     +     'exchange_2_m', 'Error in allocating memory for F_VRU_TMP!')
      mem_F_VRU_TMP = dble(4 * mem_size)
      mem_total = mem_total + mem_F_VRU_TMP
      mem_peak  = max(mem_peak, mem_total)

      call exchange_2_slab( min_xyz, max_xyz, F_VRU,
     +     nrffts_1_row, F_VRU_TMP )

      deallocate(F_VRU)
      mem_total = mem_total - mem_F_VRU

      mem_size = nrffts_1_row*(-frst_lst_u(1,iam)+frst_lst_u(2,iam)+1)
      mem_F_XK_dZ = dble(4 * (-min_x+max_x+1) * mem_size)
      allocate(F_XK_dZ((-min_x+max_x+1) * mem_size), stat = alloc_stat)
      if(alloc_stat > 0 ) then
         call error_stop (progname, subname,  
     +        ' Memory allocation failed')
      endif
      mem_total = mem_total + mem_F_XK_dZ 
      mem_peak  = max(mem_peak, mem_total)

      call rearrange_2_slab(min_xyz, max_xyz, nrffts_1_row,
     +     F_VRU_TMP, min_x, max_x, F_XK_dZ )

C     Free memory used in 2D Fourier synthesis.
      deallocate(F_VRU_TMP)
      mem_total = mem_total - mem_F_VRU_TMP
      
C ----- This is the end of exchanging data for 1D Fourier synthesis ----
C ------ Before move onto the next step, synchronize all the nodes -----

      call mpi_barrier(mpi_comm_world, mpi_err)

      call system_clock(clock2)
      ticks = clock2-clock1
      if(ticks < 0) then
         ticks = clockmax + ticks
      endif
      t_exchange_2 = dble(ticks)/dble(clockrate)

C **********************************************************************
C *                        1D Fourier synthesis                        *
C **********************************************************************

      call system_clock(clock1)
      
C     Allocate storage for 1D Fourier synthesis.
      mem_size = ((max_x-min_x+1)**2) * (-first_z+last_z+1)
      if(mem_size < 0) mem_size = 0
      allocate(rho(min_x:max_x, min_y:max_y, first_z:last_z),
     +     stat = alloc_stat)
      if (alloc_stat > 0) call error_stop(progname, subname, 
     +     'Error in allocating memory for RHO')
      mem_rho = dble(4 * mem_size)
      mem_total  = mem_total + mem_rho 
      mem_peak   = max(mem_peak, mem_total)
      
C     CALL FFTSYNTH_2
C     ***************
      call fftsynth_2_m_slab( nrffts_1_row, min_x, max_x, min_y, 
     +     max_y, min_z, max_z, F_XK_dZ, nrffts_2_row, 
     +     nrffts_2_col, nrffts_2_omega, rho, 
     +     mem_fftsynth_2_m )
      mem_peak = max(mem_peak, mem_total+mem_fftsynth_2_m)
      
C     Free memory used in 2D Fourier synthesis.
      deallocate(F_XK_dZ)
      mem_total = mem_total - mem_F_XK_dZ

C -------------- This is the end of 1D Fourier synthesis --------------
C ----- Before move onto the next step, synchronize all the nodes -----

      call mpi_barrier(mpi_comm_world, mpi_err)

      call system_clock(clock2)
      ticks = clock2-clock1
      if(ticks < 0) then
         ticks = clockmax + ticks
      endif
      t_fftsynth_2 = dble(ticks)/dble(clockrate)

C *********************************************************************
Cji   clear the density map and output density map
C *********************************************************************

      call system_clock(clock1)
      
      mem_size = (-den_min_x+den_max_x+1)*
     +     (-den_min_y+den_max_y+1)* max_z_slab
Cji   according to the set_indice, node 1 alway has the most largest slab,

      allocate(density_map(den_min_x:den_max_x, den_min_y:den_max_y,
     $     max_z_slab), stat = alloc_stat)
      mem_density_map = dble(4*mem_size)
      mem_total = mem_total + mem_density_map
      mem_peak  = max(mem_peak, mem_total)
      if (alloc_stat > 0) call error_stop(progname,
     +     subname,  'Error in allocating density_map')
      
      call density_clear (rho, min_x, max_x, min_y, max_y,
     +     den_min_x, den_max_x,
     +     frst_lst_z(1,iam), frst_lst_z(2,iam),
     +     density_map, map_min_local, map_max_local)
      
C     Get the global maximum and minimum density over entire map
      call mpi_reduce(map_min_local, map_min, 1, MPI_REAL, MPI_MIN,
     +     0, mpi_comm_world, mpi_err)
      call mpi_reduce(map_max_local, map_max, 1, MPI_REAL, MPI_MAX,
     +     0, mpi_comm_world, mpi_err)
      if( mpi_err .ne. 0 ) call error_stop(progname,'density_clear',
     +     'Error in calling MPI_BCAST background')

      deallocate(rho)
      mem_total = mem_total - mem_rho

Cji   Write density map into MAP file.

      call output_density( density_map, den_min_x,
     +     den_max_x, max_z_slab, MAP_MIN, MAP_MAX )

C     Free memory for storing the density values.

      deallocate(density_map)
      mem_total = mem_total - mem_density_map

C ------- This the end of writing density values into MAP file --------
C ----- Before reporting the timing the memory allocating mesage, -----
C ------------------- synchronize all the nodes. ----------------------

      call mpi_barrier(mpi_comm_world, mpi_err)

      call system_clock(clock2)
      ticks = clock2-clock1
      if(ticks < 0) then
         ticks = clockmax + ticks
      endif
      t_write_output = dble(ticks)/dble(clockrate)

C *********************************************************************
C *       Reporting the timing and memory allocating messages         *
C *********************************************************************

      if (mem_total .ne. 0) print *,'mem_total=',mem_total,
     +     'At the end, mem_total should be 0!!!'

      if ( iam == 0 ) then
C     Print timing message.
         t_seq = t_init + t_analy + t_write_output
         t_comm = t_exchange_2 + t_gather
         t_par = t_interp  + t_fftanaly + 
     +        t_solvesys + t_fftsynth_1 + t_fftsynth_2
         time = t_seq + t_comm + t_par
         
         write(*,40) iam, t_init, t_analy, t_fftanaly, t_interp,
     $        t_solvesys, t_fftsynth_1, t_exchange_2, t_fftsynth_2,
     $        t_gather, t_write_output, time, t_seq, t_comm, t_par

 40      format (/ 'Node ', i3, 
     +        ' Time used in initialization                      =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in read images                         =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in 2D FFT analysis                     =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in 2D interpolation                    =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in solving linear systems              =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in 2-D Fourier synthesis               =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in exchanging data before fftsynth_2   =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in 1-D Fourier synthesis               =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in exchanging data before output       =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in writing output                      =', 
     +        f15.4, ' ' / 8x, 
     +        ' Total time used                                  =',
     +        f15.4, ' ' / 8x,
     +        '     (Seq, Com, Par):'/   
     +        '    (', f15.4, ',', f15.4, ',', f15.4, ')')

C     Print memory allocating message.
         write(*,60) iam, mem_peak,
     +        mem_interpolants, mem_f_hkl, mem_f_hkl_ref, mem_f_vru,
     +        mem_F_XK_dZ, mem_rho, mem_density_map,mem_read_files,
     +        mem_cmpt_intrps, mem_solvesys_m, mem_fftsynth_1_m,
     +        mem_exchange_2_m, mem_fftsynth_2_m
 60      format(/ 'Node ', i3, ' Memory allocation in BYTES:', /
     +        8x, ' Peak memory allocated   =', f18.1 /
     +        8x, ' Memory for INTERPOLANTS =', f18.1 /
     +        8x, ' Memory for F_HKL        =', f18.1 /
     +        8x, ' Memory for f_hkl_ref    =', f18.1 /
     +        8x, ' Memory for F_VRU        =', f18.1 /
     +        8x, ' Memory for F_XK_dZ      =', f18.1 /
     +        8x, ' Memory for RHO          =', f18.1 /
     +        8x, ' Memory for density_map  =', f18.1 /
     +        8x, ' Memory used in READ_FILES   =', f18.1 /
     +        8x, ' Memory used in CMPT_INTRPS  =', f18.1 /
     +        8x, ' Memory used in SOLVESYS_M   =', f18.1 /
     +        8x, ' Memory used in FFTSYNTH_1_M =', f18.1 /
     +        8x, ' Memory used in EXCHANGE_2_M =', f18.1 /
     +        8x, ' Memory used in FFTSYNTH_2_M =', f18.1 )
         
         print *, '**********finished nicely**********'
      endif

      call MPI_FINALIZE ( mpi_err )

      stop
      end program p3dr
