      subroutine cut_map(map3d, max_slab, in_rad, out_rad, cut_step,
     $     cut_weight, npix_col, npix_row, in_3d_name, out_file_name)

C     Read in a 3D map, apply the following transformation, and write
C     out the new map
C     
C                           r < (in_rad - cut_step)  --> map3d = 0
C     (in_rad - cut_step) < r < in_rad               --> map3d = map3d*G(d)
C                  in_rad < r < out_rad              --> map3d unchanged
C                 out_rad < r < (out_rad + cut_step) --> map3d = map3d*G(d)
C    (out_rad + cut_step) < r                        --> map3d = 0
C
C     in_rad and out_rad specify the radii of the annulus containing the
C     ordered portion of the map, cut_step is the distance over which to
C     apply a Gaussian falloff, d is the distance between the location
C     of the map point and either the inner radius (r < in_rad) or the
C     other radius (r > out_rad), and G(d) is the Gaussian function
C     given by
C
C     G(d) = exp(-d**2 / (-cut_step**2/log(cut_weight)))

      implicit none
      
      include 'mpif.h'
      include '../include/allprogs.inc'
      INCLUDE '../include/pifhead.inc'
      
      type (PIFHEAD_T) MAPHEAD, IMGHEAD
      COMMON /PIFHEAD/   MAPHEAD, IMGHEAD

      integer, intent(in) :: in_rad, out_rad, cut_step, max_slab
      integer, intent(in) :: npix_col, npix_row
      real, intent(in)    ::  cut_weight
      character(len=257), intent(in) :: in_3d_name, out_file_name
      real, intent(out)   ::  map3D(npix_col, npix_row, max_slab)

      integer i, j, f_z, l_z, NCOL, NROW, NSEC, x, y, z, sec, outmode
      integer stat(mpi_status_size)
      character(len=7) :: subname = 'Cut_Map'
      real    radius, dist
      double precision factor

C     Root process reads in the global and data headers from the current
C     map and writes information to new file. It then reads in chunks of
C     the map and distributes to other processes.

      if (iam == 0) then
         
C     Read header
         MAPHEAD%LUNUM  = 1
         CALL PIF_OPEN (in_3d_name, MAPHEAD%LUNUM, 'r')
         CALL PIF_READ_GH (MAPHEAD)
         outmode = MAPHEAD%MODE
         IF(outmode.NE.20 .and. outmode.NE.40)
     +        call error_stop(progname,'read_map',
     +        'Error in 3D density map mode, it should be 20 or 40')
         
C     Write global header
         MAPHEAD%LUNUM  = 8
         CALL PIF_OPEN (out_file_name, MAPHEAD%LUNUM, 'w')
         CALL PIF_WRITE_GH (MAPHEAD)
         
         MAPHEAD%LUNUM  = 1
         MAPHEAD%IMGNUM = 1     ! LOAD UP "IMG" #1
         CALL PIF_READ_DH (MAPHEAD)
         NCOL = MAPHEAD%NUMCOLS
         NROW = MAPHEAD%NUMROWS
         NSEC = MAPHEAD%NUMSECS
         if(NCOL.ne.npix_col.or.NROW.ne.npix_row.or.NSEC.ne.npix_col)
     +        call error_stop(progname,subname,
     +        'Density map COL,ROW,SEC not consistent with image size')

C     Write data header
         MAPHEAD%LUNUM  = 8
         MAPHEAD%IMGNUM = 1     ! LOAD UP "IMG" #1
         CALL PIF_WRITE_DH (MAPHEAD)

C     Read in chunks of the map and distribute to other processes
         MAPHEAD%LUNUM  = 1
         do i = num_nodes-1, 1, -1
            f_z = frst_lst_y(1,i)
            l_z = frst_lst_y(2,i)

            if (outmode == 20) then
               call PIF_READ_MAPI2(MAPHEAD,map3D,f_z,l_z,NCOL,NROW,NSEC)
            elseif (outmode == 40) then
               call PIF_READ_MAPI4(MAPHEAD,map3D,f_z,l_z,NCOL,NROW,NSEC)
            endif

            call MPI_SEND(map3D, (l_z - f_z + 1)*NCOL*NROW,
     +           mpi_real, i, i, mpi_comm_world, mpi_err )
            if (mpi_err .ne. 0) call error_stop(progname, subname,
     +           'Terminated because of mpi_send error')
         enddo

C     Read in the last chunk and keep for self
         f_z = frst_lst_y(1,iam)
         l_z = frst_lst_y(2,iam)
         if (outmode == 20) then
            call PIF_READ_MAPI2(MAPHEAD,map3D,f_z,l_z,NCOL,NROW,NSEC)
         elseif (outmode == 40) then
            call PIF_READ_MAPI4(MAPHEAD,map3D,f_z,l_z,NCOL,NROW,NSEC)
         endif
         CALL PIF_CLOSE (1)

      else 

C     Receive chunks of map from root
         f_z = frst_lst_y(1,iam)
         l_z = frst_lst_y(2,iam)
         call MPI_RECV( map3D,(l_z - f_z + 1)*Npix_COL*Npix_ROW,
     +        mpi_real, 0, iam, mpi_comm_world, stat, mpi_err)
         if (mpi_err.ne.0) call error_stop(progname, subname,
     +        'Terminated because of mpi_recv error')
      endif

C     Apply transformation to map
      factor = - cut_step**2 / log(cut_weight)
      do z = 1, l_z - f_z + 1
         do y = 1, npix_row
            do x = 1, npix_col 
               radius = (x - (npix_col+1)/2)**2 + 
     +              (y - (npix_row+1)/2)**2
     +              + (z + f_z - 1 - (npix_col+1)/2)**2
               radius = sqrt(radius)
               if (radius > real(out_rad+cut_step)) then
                  map3D(x,y,z) = 0.0
               elseif (radius < real(in_rad-cut_step)) then
                  map3D(x,y,z) = 0.0
               elseif (radius > real(out_rad)) then
                  dist = radius - out_rad
                  map3D(x,y,z) = map3D(x,y,z) * exp( - dist**2/factor)
               elseif (radius < real(in_rad)) then
                  dist = in_rad - radius
                  map3D(x,y,z) = map3D(x,y,z) * exp( - dist**2/factor)
               endif
            enddo
         enddo
      enddo
      
C     Root process collects chunks of transformed map from other
C     processes and writes out the new map
       
      if (iam == 0) then
         MAPHEAD%LUNUM  = 8
         SEC = 1
         do i = 1, l_z - f_z + 1
            if (outmode == 40) then
               CALL PIF_WRITE_SECI4 (MAPHEAD,map3D(1,1,i),NCOL,NROW,SEC)
            elseif (outmode == 20) then
               CALL PIF_WRITE_SECI2 (MAPHEAD,map3D(1,1,i),NCOL,NROW,SEC)
            endif
            sec = sec +1
         enddo
         
         do i = 1, num_nodes-1
            f_z = frst_lst_y(1,i)
            l_z = frst_lst_y(2,i)
            call MPI_RECV(map3D(1,1,1), (l_z - f_z + 1)* (NCOL * NROW),
     +           mpi_real, i, i, mpi_comm_world, stat, mpi_err)
            do j = 1, l_z - f_z + 1
               if (outmode == 40) then
                  CALL PIF_WRITE_SECI4 (MAPHEAD,map3D(1,1,j),
     +                 NCOL,NROW,SEC)
               elseif (outmode == 20) then
                  CALL PIF_WRITE_SECI2 (MAPHEAD,map3D(1,1,j),
     +                 NCOL,NROW,SEC)
               endif
               sec = sec +1
            enddo
         enddo
         CALL PIF_CLOSE (MAPHEAD%LUNUM)
      else
         call MPI_SEND(map3D(1,1,1),
     +        (l_z - f_z + 1) * (npix_col*npix_row),
     +        mpi_real, 0, iam, mpi_comm_world, mpi_err )
         if (mpi_err.ne.0) call error_stop(progname, subname,
     +        'Terminated because of  mpi_send  error')
      endif
      
      return
      end subroutine cut_map
